use std::{
    sync::{mpsc, Arc, Mutex},
    thread,
};

type Job = Box<dyn FnOnce() + Send + 'static>;

pub struct ThreadPool {
    sender: Option<mpsc::Sender<Job>>,
    workers: Vec<Worker>,
}

impl ThreadPool {
    /// Create a new ThreadPool
    ///
    /// size: Number of threads to spawn
    pub fn build(size: usize) -> Option<Self> {
        let (sender, receiver) = mpsc::channel();
        // Wrap the receiver in an Arc and Mutex so we can share it among threads and only have one user at a time
        let receiver = Arc::new(Mutex::new(receiver));

        let mut workers = Vec::with_capacity(size);
        for id in 0..size {
            match Worker::build(id, receiver.clone()) {
                Some(worker) => workers.push(worker),
                None => return None,
            }
        }

        Some(ThreadPool {
            sender: Some(sender),
            workers,
        })
    }

    /// Pass a function to a worker thread to execute
    ///
    /// f: Function to execute
    pub fn execute<F>(&self, f: F) -> Result<(), mpsc::SendError<Job>>
    where
        F: FnOnce() + Send + 'static,
    {
        let job = Box::new(f);

        match self.sender.as_ref() {
            Some(sender) => sender.send(job),
            None => Err(mpsc::SendError(job)),
        }
    }
}

impl Drop for ThreadPool {
    fn drop(&mut self) {
        drop(self.sender.take());

        for worker in &mut self.workers {
            println!("Shutting down worker {}", worker.id);
            if let Some(thread) = worker.thread.take() {
                thread
                    .join()
                    .expect("Failed to join worker thread on shutdown!")
            }
        }
    }
}

struct Worker {
    id: usize,
    thread: Option<thread::JoinHandle<()>>,
}

impl Worker {
    /// Create a new Worker
    ///
    /// id: Worker ID
    /// receiver: Receiver to listen for jobs, wrapped in an Arc and Mutex to share among threads
    fn build(id: usize, receiver: Arc<Mutex<mpsc::Receiver<Job>>>) -> Option<Self> {
        let builder = thread::Builder::new().name(format!("worker-{}", id));

        let thread = match builder.spawn(move || {
            println!("Worker {} spawned", id);
            loop {
                let job = match receiver
                    .lock()
                    .expect("Failed to lock receiver! Maybe the Mutex is poisoned?")
                    .recv()
                {
                    Ok(job) => job,
                    Err(_) => break,
                };

                job();
            }
        }) {
            Ok(thread) => Some(thread),
            Err(_) => return None,
        };

        Some(Worker { id, thread })
    }
}
